from abc import ABC, abstractmethod
from typing import List, TypeVar

from .exceptions import AbortRetries, FailuresAfterRetrying
from .steps import Step


InputState = TypeVar("InputState")
OutputState = TypeVar("OutputState")


class StepWithRetries(Step[InputState, OutputState], ABC):
    @abstractmethod
    def retry(self, state: InputState, failures: List[Exception]) -> OutputState:
        pass


def attempt_retries(step: Step, times: int) -> StepWithRetries:
    class AutoRetryStep(StepWithRetries):
        def __init__(self, wrapped_step: Step, max_tries: int) -> None:
            self.wrapped_step = wrapped_step
            self.attempts_made = 0
            self.max_tries = max_tries

        def compensate(self, state) -> None:
            self.wrapped_step.compensate(state)

        def execute(self, state):
            self.attempts_made += 1
            return self.wrapped_step.execute(state)

        def retry(self, state, _failures):
            if self.attempts_made > self.max_tries:
                raise AbortRetries("Maximum number of retries hit")
            return self.execute(state)

    return AutoRetryStep(step, times)


def execute_step_retry(state, step: StepWithRetries, previous_errors: List[Exception]):
    try:
        return step.retry(state, previous_errors)
    except AbortRetries as _give_up:
        raise FailuresAfterRetrying(previous_errors)
    except Exception as e:
        return execute_step_retry(state, step, previous_errors + [e])



import inspect
from typing import Any, Callable, Tuple, TypeVar

X = TypeVar("X")
FunctionPair = Tuple[Callable[[Any], X], Callable[[X], Any]]
PlainStateChangingFunction = Callable[[Any], Any]


def arity(func: Callable) -> int:
    return len(inspect.signature(func).parameters)


def is_arity_one_pair(thing: Any) -> bool:
    return (
        hasattr(thing, "__len__")
        and len(thing) == 2
        and callable(thing[0])
        and arity(thing[0]) == 1
        and callable(thing[1])
        and arity(thing[1]) == 1
    )



from typing import Tuple, Any, List, Iterable

from .exceptions import CompensationFailure
from .retries import StepWithRetries, execute_step_retry
from .steps import Step, build_step_list, StepLike


def _compensate_completed_steps(completed_steps: List[Tuple[Step, Any]]):
    failures = []
    for (step, state) in reversed(completed_steps):
        try:
            step.compensate(state)
        except Exception as failure:
            failures.append(failure)
    if failures != []:
        raise CompensationFailure(failures)


def _execute_step(state, step: Step):
    try:
        return step.execute(state)
    except Exception as e:
        if isinstance(step, StepWithRetries):
            return execute_step_retry(state, step, [e])
        raise e


def run_transaction(steps: Iterable[StepLike], starting_state=None):
    steps = build_step_list(steps)
    completed_steps: List[Tuple[Step, Any]] = []
    state = starting_state
    try:
        for step in steps:
            state = _execute_step(state, step)
            completed_steps.append((step, state))
        return state

    except Exception as error:
        _compensate_completed_steps(completed_steps)
        raise error
